import json
import logging
import sqlite3
from gen3.auth import Gen3Auth
from anvil.clients.gen3_auth import Gen3TerraAuth
from anvil.clients.gen3_auth import TERRA_TOKEN_URL
from gen3.query import Gen3Query
from tabulate import tabulate

from anvil.etl.utilities.entities import dict_factory


def drs_extractor(gen3_credentials_path, output_path, use_terra_credentials, expected_row_count):
    """Retrieve DRS url from Gen3's flat file index."""
    # TODO - consider wrapping this using the Entities class
    gen3_endpoint = "https://gen3.theanvil.io"

    # Install n API Key downloaded from the
    # commons' "Profile" page at ~/.gen3/credentials.json

    if use_terra_credentials:
        auth = Gen3TerraAuth(endpoint=gen3_endpoint, terra_auth_url=TERRA_TOKEN_URL, user_email=None)
    else:
        auth = Gen3Auth(endpoint=gen3_endpoint, refresh_file=gen3_credentials_path)

    logger = logging.getLogger(__name__)

    query_client = Gen3Query(auth)
    logger.info('Starting export of data from gen3.')
    raw_data = query_client.raw_data_download(data_type='file', fields='node_id,project_id,anvil_project_id,subject_submitter_id,sample_submitter_id,sequencing_assay_submitter_id,file_name,file_size,md5sum,submitter_id,md5sum,drs_id,_subject_id'.split(','))

    assert len(raw_data) > expected_row_count, f"Expected over {expected_row_count} file records, got {len(raw_data)} instead.  Projects {set(sorted([(r['project_id']) for r in raw_data]))}"

    logger.info(f"retrieved {len(raw_data)} file records from gen3.")

    sqlite_path = f'{output_path}/drs_file.sqlite'
    _conn = sqlite3.connect(sqlite_path)
    cur = _conn.cursor()
    cur.executescript("""
    CREATE TABLE IF NOT EXISTS drs_file (
        md5sum text,
        sequencing_id text PRIMARY KEY,
        file_name text,
        ga4gh_drs_uri text,
        sample_submitter_id text,
        subject_submitter_id text,
        subject_id text,
        project_id text,
        anvil_project_id text
    );
    """)
    _conn.commit()
    # optimize for single thread speed
    _conn.execute('PRAGMA synchronous = OFF')
    _conn.execute('PRAGMA journal_mode = OFF')
    _conn.commit()
    _conn.close()

    _conn = sqlite3.connect(sqlite_path, check_same_thread=False, isolation_level='DEFERRED')
    cur = _conn.cursor()
    logger.info(f'Starting import of data into sqlite {sqlite_path}.')

    def _first(_array):
        """Return first element in array, or none if empty."""
        if not _array or len(_array) == 0:
            return None
        return _array[0]

    commit_threshold = 1000
    c = 0
    for row in raw_data:
        try:
            cur.execute(
                "INSERT into drs_file values (?, ?, ?, ?, ?, ?, ?, ?, ?);", (
                    row['md5sum'],
                    row['node_id'],
                    row['file_name'],
                    row['drs_id'],
                    _first(row['sample_submitter_id']),
                    _first(row['subject_submitter_id']),
                    _first(row['_subject_id']),
                    row['project_id'],
                    _first(row['anvil_project_id']),
                )
            )
        except Exception:
            logger.error(row)
            logger.error("Fatal error writing row above to sqlite.", exc_info=True)
            break
        c += 1
        if c > commit_threshold:
            _conn.commit()
            c = 0
    _conn.commit()

    logger.info('Indexing')
    cur.executescript("""
    CREATE INDEX IF NOT EXISTS drs_file_md5sum ON drs_file(md5sum);
    CREATE  INDEX IF NOT EXISTS drs_file_file_name ON drs_file(file_name);
    """)
    _conn.commit()
    logger.info(f'Created {sqlite_path}')

    def _dict_factory(cursor, row):
        d = {}
        for idx, col in enumerate(cursor.description):
            d[col[0]] = row[idx]
        return d

    _conn = sqlite3.connect(sqlite_path)
    _conn.row_factory = _dict_factory
    cur = _conn.cursor()

    dataset = cur.execute('SELECT project_id as "gen3_project_id", anvil_project_id, count(*) as "file_count" FROM drs_file group by  project_id, anvil_project_id').fetchall()

    header = dataset[0].keys()
    rows = [x.values() for x in dataset]
    logger.info(f"\nExtracted File Counts\n{tabulate(rows, header)}")


class DRSReader:
    """Read items in sqlite."""

    def __init__(self, output_path):
        """Set up sqlite db."""
        self._path = f'{output_path}/drs_file.sqlite'
        self._conn = sqlite3.connect(self._path, check_same_thread=True)
        self._conn.row_factory = dict_factory

    def get(self, file_name=None, md5sum=None):
        """Retrieve an item."""
        assert file_name or md5sum, "Please provide either file_name or md5sum"
        cur = self._conn.cursor()
        data = cur.execute("SELECT * FROM drs_file where file_name=? or md5sum=?", (file_name, md5sum, )).fetchone()
        cur.close()
        return data



import click
import logging

from anvil.etl.utilities.shell_helper import run_cmd

logger = logging.getLogger(__name__)


@click.group(name='data-set')
@click.pass_context
def data_set_cli(ctx):
    pass  # dummy


@data_set_cli.command(name='create')
@click.option('--data_set',  envvar='GOOGLE_DATASET', help='data set name', show_default=True)
@click.pass_context
def create_dataset(ctx, data_set):
    """Create dataset."""
    check_script = """
        gcloud healthcare datasets list --location=$GOOGLE_LOCATION | grep $GOOGLE_DATASET
    """
    check_results = run_cmd(check_script)
    if check_results and data_set in check_results:
        logger.info(f"{data_set} already exists")
        return
    create_script = """
        gcloud healthcare datasets create $GOOGLE_DATASET --location=$GOOGLE_LOCATION
    """
    print(run_cmd(create_script))


@data_set_cli.command(name='delete')
@click.option('--data_set',  envvar='GOOGLE_DATASET', help='data set name', show_default=True)
@click.pass_context
def delete_dataset(ctx, data_set):
    """Delete dataset."""
    rm_script = """
        gcloud healthcare datasets delete $GOOGLE_DATASET --location=$GOOGLE_LOCATION  --quiet
    """
    print(run_cmd(rm_script))



import os
import json
import requests

REST_URL = "http://data.bioontology.org"

def _get_json(url, api_key):
    response = requests.get(url, headers={'Authorization': f'apikey token={api_key}'})
    response.raise_for_status()
    assert isinstance(response.json(), dict), f"{url} returned {response.text} ?"
    return response.json()


def lookup_term(term, api_key, ontology_preferences=['MONDO', 'DOID']):
    """Generator, return classes that match ontology_preferences."""
    assert api_key, "Please provide an API key.  See https://bioportal.bioontology.org/help#Getting_an_API_key"
    url = f"{REST_URL}/search?q={term}"
    results = _get_json(url, api_key)
    from pprint import pprint
    for ontology_preference in ontology_preferences:
        for item in results['collection']:
            if ontology_preference in item['@id']:
                yield (ontology_preference, item)





import click
import logging


from anvil.etl import DEFAULT_GEN3_CREDENTIALS_PATH, DEFAULT_GOOGLE_PROJECT

from anvil.etl.extractors.data_ingestion_tracker import write as write_projects
from anvil.etl.extractors.gen3 import drs_extractor
from anvil.etl.extractors.google import extract_buckets
from anvil.etl.extractors.terra import cli as terra_cli, logger as terra_logger


logger = logging.getLogger(__name__)


@click.group(name='extract')
@click.pass_context
def extract(ctx):
    """ETL meta data from sources."""
    logger.setLevel(ctx.obj['log_level'])
    terra_logger.setLevel(ctx.obj['log_level'])


@extract.command('google')
@click.option('--user_project', default=DEFAULT_GOOGLE_PROJECT, help="AnVIL buckets use the `Requester Pays` feature. Please include a billing project. Defaults to GOOGLE_PROJECT")
@click.pass_context
def _google(ctx, user_project):
    """Extract bucket data, write to db."""
    extract_buckets(ctx.obj['output_path'], user_project)


@extract.command('spreadsheet')
@click.pass_context
def _spreadsheet(ctx):
    """Extract data_ingestion_tracker spreadsheet from anvil portal repo."""
    write_projects(ctx.obj['output_path'])


@extract.command('gen3')
@click.option('--gen3_credentials_path', default=DEFAULT_GEN3_CREDENTIALS_PATH, help='Gen3 credentials. https://anvilproject.org/learn/introduction/getting-started-with-gen3#profile-page---api-keys-and-project-access', show_default=True)
@click.option('--use_terra_credentials', is_flag=True, default=True, help='Running in terra VM, use terra authenticator to access gen3.', show_default=True)
@click.option('--expected_row_count', default=175000, help="Minimum number of file records expected.", show_default=True)
@click.pass_context
def _gen3(ctx, gen3_credentials_path, use_terra_credentials, expected_row_count):
    """Extract meta data from gen3, including drs identifiers."""
    drs_extractor(output_path=ctx.obj['output_path'], expected_row_count=expected_row_count, gen3_credentials_path=gen3_credentials_path, use_terra_credentials=use_terra_credentials)


@extract.command('clean')
@click.pass_context
def _clean(ctx):
    """Drop all databases and downloads."""
    logger.info(f"TODO: {ctx.command.name} remove files from {ctx.obj['output_path']}")


# TODO make terra a subgroup and incorporate commands from extractors.terra
@extract.group(name='terra')
@click.pass_context
def terra(ctx):
    """Commands to download and analyze terra workspaces."""
    pass


@terra.command('fetch')
@click.pass_context
def _terra_fetch(ctx):
    """Extract terra workspaces, write to db."""
    logger.info(f"TODO: {ctx.command.name} write to {ctx.obj['output_path']}")


for command in terra_cli.commands.values():
    terra.add_command(command)



import click
import logging


from anvil.etl.loaders.fhir.initializer import fhir as fhir_cli, logger as fhir_logger
from anvil.etl.loaders.fhir.dataset import data_set_cli, logger as data_set_logger
from anvil.etl.loaders.fhir.datastore import data_store_cli, logger as data_store_logger
from anvil.etl.loaders.fhir.implementation_guide import ig_cli, logger as ig_logger


logger = logging.getLogger(__name__)


@click.group(name='load')
@click.pass_context
def load(ctx):
    """Load commands."""
    logger.setLevel(ctx.obj['log_level'])
    fhir_logger.setLevel(ctx.obj['log_level'])
    data_set_logger.setLevel(ctx.obj['log_level'])
    data_store_logger.setLevel(ctx.obj['log_level'])
    ig_logger.setLevel(ctx.obj['log_level'])


@load.command(name='pfb')
@click.pass_context
def _pfb(ctx):
    """Write FHIR schema and data to PFB"""
    logger.info(f"TODO: {ctx.command.name} write files from {ctx.obj['output_path']} to PFB")


@load.group(name='fhir')
@click.pass_context
def fhir(ctx):
    """Commands to setup and load fhir server."""
    pass


@fhir.group(name='data-set')
@click.pass_context
def data_set(ctx):
    """Commands to create and delete data_set."""
    pass


@fhir.group(name='data-store')
@click.pass_context
def data_store(ctx):
    """Commands to create and delete data_store."""
    pass


@fhir.group(name='IG')
@click.pass_context
def ig(ctx):
    """Commands to create and delete implementation guide."""
    pass


for command in fhir_cli.commands.values():
    fhir.add_command(command)

for command in data_set_cli.commands.values():
    data_set.add_command(command)

for command in data_store_cli.commands.values():
    data_store.add_command(command)

for command in ig_cli.commands.values():
    ig.add_command(command)



import logging
import sys
from collections import defaultdict


class LogCapture(logging.Logger):
    """Capture warnings and errors in `entries`, assumes tuple is logged, entries.key = tuple[0]."""
    def __init__(self, name='anvil-etl-transformers'):
        self.entries = defaultdict(list)
        super(LogCapture, self).__init__(name=name)

    def _persist(self, msg):
        self.entries[msg[0]].append(msg)

    def error(self, msg, *args, **kwargs):
        super().error(msg, *args, **kwargs)
        self._persist(msg)

    def warning(self, msg, *args, **kwargs):
        super().warning(msg, *args, **kwargs)
        self._persist(msg)

    def clear(self):
        self.entries = defaultdict(list)


logger = LogCapture()
if len(logging.root.handlers) == 0:
    print("Please initialize logging.",  file=sys.stderr)
else:
    logger.addHandler(logging.root.handlers[0])


def _recursive_default_dict():
    """Recursive default dict, any key defaults to a dict."""
    return defaultdict(_recursive_default_dict)



"""Client for transforming workspaces."""

import logging


class Transformer(object):
    """Render workspace into target form, Subclass and override transform_* per your requirements."""

    def __init__(self, *args, workspace):
        """Initialize instance."""
        self._logger = logging.getLogger(__name__)
        self.workspace = workspace

    def transform(self):
        """Transform entities."""
        try:
            for w in self.transform_workspace(self.workspace):
                yield w
            for subject in self.workspace.subjects:
                for s in self.transform_subject(subject):
                    yield s
                    for sample in subject.samples:
                        for s in self.transform_sample(sample, subject):
                            yield s
        except Exception as e:
            logging.getLogger(__name__).warning(f"{self.workspace.id} {e}")

    def transform_workspace(self, workspace):
        """Transform workspace (noop)."""
        yield workspace

    def transform_subject(self, subject):
        """Transform subject (noop)."""
        yield subject

    def transform_sample(self, sample):
        """Transform sample (noop)."""
        yield sample

    def transform_blob(self, blob):
        """Transform blob (noop)."""
        yield blob



#!/usr/bin/env python3

"""Extract AnVIL Data Ingestion Tracker spreadsheet."""

import json
import logging
import requests
from io import StringIO
import csv

logging.basicConfig(level=logging.INFO, format='%(asctime)s %(filename)s %(levelname)-8s %(message)s')
logger = logging.getLogger(__name__)

URL = 'https://raw.githubusercontent.com/anvilproject/anvil-portal/main/plugins/utils/dashboard-source-anvil.tsv'


def download_projects():
    """Read projects from master spreadsheet. Call `projects = [project for project in download_projects()]`"""

    response = requests.get(URL)
    response.raise_for_status()

    list_of_lists = []
    with StringIO(response.text) as fd:
        rd = csv.reader(fd, delimiter="\t", quotechar='"')
        for row in rd:
            list_of_lists.append(row)

    logger.debug(f"fetched {len(list_of_lists)} rows")

    # table headers
    keys = list_of_lists.pop(0)

    def normalize_value(v):
        """Fix the value."""
        if isinstance(v, str):
            v = v.strip()
        if v in ['None', '', 'NA', '--', 'Unspecified', None]:
            return None
        if 'no' == str(v).lower():
            return False
        if 'yes' == str(v).lower():
            return True
        if str(v).isnumeric():
            return int(v)
        return v

    for lst in list_of_lists:
        project = {keys[i]: normalize_value(v) for i, v in enumerate(lst)}
        if project['name']:
            yield {keys[i]: normalize_value(v) for i, v in enumerate(lst)}


def data_ingestion_tracker(output_path):
    """Read spreadsheet, write to json file."""
    abbreviated_projects = [
        {'workspace_name': p['name'],
         'study_accession': p['phsId'],
         'dataUseRestriction': p['library:dataUseRestriction'],
         'indication': p['library:indication']} for p in list(download_projects())]
    with open(output_path, 'w') as fp:
        json.dump(abbreviated_projects, fp)
    logger.info(f"Read {len(abbreviated_projects)} projects from {URL}. Wrote to {output_path}")





"""Represent fhir entity."""


from anvil.transformers.fhir import join, make_identifier
import logging

from anvil.transformers.fhir.organization import Organization

PRACTITIONERS = []

class Practitioner:
    """Create fhir entity."""

    class_name = "practitioner"
    resource_type = "Practitioner"

    @staticmethod
    def slug(workspace):
        """Make id."""
        investigator_name = "Unknown"
        if hasattr(workspace, 'investigator'):
            if workspace.investigator:
                investigator_name = workspace.investigator
        return make_identifier(investigator_name)

    @staticmethod
    def build_entity(workspace):
        """Create fhir entity."""
        practitioner_id = Practitioner.slug(workspace)
        investigator_name = workspace.investigator
        institute = workspace.institute
        if practitioner_id in PRACTITIONERS:
            return None
        PRACTITIONERS.append(practitioner_id)
        if not institute:
            logging.getLogger(__name__).warning(f'workspace {workspace.id} missing institute')
        entity = {
            "resourceType": Practitioner.resource_type,
            "id": practitioner_id,
            "meta": {
                "profile": [
                    "http://hl7.org/fhir/StructureDefinition/Practitioner"
                ]
            },
            "identifier": [
                {
                    "system": f"https://anvil.terra.bio/#workspaces/anvil-datastorage/{workspace.attributes.workspace.name}/data",
                    "value": investigator_name,
                },
                {
                    "system": "urn:ncpi:unique-string",
                    "value": join(Practitioner.resource_type, practitioner_id),
                }
            ],
            "name": [{"text": investigator_name}],
        }

        return entity

    @staticmethod
    def build_practitioner_role(workspace):
        """Create fhir entity."""
        practitioner_id = Practitioner.slug(workspace)
        organization_id = Organization.slug(workspace)
        id = f"{organization_id}-{practitioner_id}"
        
        if id in PRACTITIONERS:
            return None

        PRACTITIONERS.append(id)            

        entity = {
            "resourceType": 'PractitionerRole',
            "id": f"{id}",
            "meta": {
                "profile": [
                    "http://hl7.org/fhir/StructureDefinition/PractitionerRole"
                ]
            },
            "identifier": [
                {
                    "system": f"https://anvil.terra.bio",
                    "value": id,
                }
            ],
            "practitioner": {"reference": f"Practitioner/{practitioner_id}"},
            "organization": {"reference": f"Organization/{organization_id}"}
        }
        return entity        


"""Represent fhir entity."""

from anvil.transformers.fhir import CANONICAL
from anvil.transformers.fhir.research_study import ResearchStudy

logged_already = []


class ResearchStudyObservation:
    """Create fhir entity."""

    class_name = "observation"
    resource_type = "Observation"

    @staticmethod
    def build_entity(workspace):
        """Create FHIR entity."""
        slug = ResearchStudy.slug(workspace)

        entity = {
            "resourceType": "Observation",
            "id": slug,
            "status": "final",
            "code": {
                "coding": [
                    {
                        "system": f"{CANONICAL}",
                        "code": "Summary",
                        "display": "Variable Summary"
                    }
                ]
            },
            "focus": [
                {
                    "reference": f"ResearchStudy/{slug}"
                }
            ],
            "component": [
                # {
                # "code": {
                #     "coding": [
                #     {
                #         "code": "CohortCount",
                #         "display": "Number of Cohorts"
                #     }
                #     ]
                # },
                # "valueInteger": 41
                # },
                {
                    "code": {
                        "coding": [
                            {
                                "system": f"{CANONICAL}",
                                "code": "SampleCount",
                                "display": "Number of Samples"
                            }
                        ]
                    },
                    "valueInteger": len(workspace.samples)
                },
                {
                    "code": {
                        "coding": [
                            {
                                "system": f"{CANONICAL}",
                                "code": "Participant",
                                "display": "Number of Participants"
                            }
                        ]
                    },
                    "valueInteger": len(workspace.subjects)
                },
                {
                    "code": {
                        "coding": [
                            {
                                "system": f"{CANONICAL}",
                                "code": "StorageSize",
                                "display": "Size on Disk"
                            }
                        ]
                    },
                    "valueQuantity": {
                        "value": sum(workspace.blob_sizes.values()),
                        "system": "http://unitsofmeasure.org",
                        "code": "L"
                    }
                }
            ]
        }
        return entity



# -*- coding: utf-8 -*-
"""Google gcloud access_token handling class for smart-on-fhir/client-py FHIR client."""
import logging
from subprocess import Popen, PIPE

from fhirclient import auth

logger = logging.getLogger(__name__)


class FHIRAuthError(Exception):
    """Reports any problem retrieving access token from gcloud."""

    pass


class GoogleFHIRAuth(auth.FHIRAuth):
    """Google gcloud access_token handling class for smart-on-fhir/client-py FHIR client.

    Requires:
        `pip install -e git+https://github.com/smart-on-fhir/client-py#egg=fhirclient`

    :param access_token: Optional access token, if none provided `gcloud auth print-access-token` is used

    Examples:
        from fhirclient import client
        from anvil.fhir.smart_auth import GoogleFHIRAuth
        settings = {
            'app_id': 'my_web_app',
            'api_base': 'https://healthcare.googleapis.com/v1/projects/gcp-testing-308520/locations/us-east4/datasets/testset/fhirStores/fhirstore/fhir'
        }
        smart = client.FHIRClient(settings=settings)
        # optionally pass token
        smart.server.auth = GoogleFHIRAuth(access_token=''ya29.abcd...')
        smart.prepare()
        assert smart.ready, "server should be ready"
        # search for all ResearchStudy
        import fhirclient.models.researchstudy as rs
        [s.title for s in rs.ResearchStudy.where(struct={}).perform_resources(smart.server)]
        >>>
        ['1000g-high-coverage-2019', 'my NCPI research study example']
    """

    auth_type = 'bearer'

    def __init__(self, state=None, access_token=None):
        """Initialize access_token, call super."""
        self.access_token = access_token
        if not self.access_token:
            self.access_token = self._get_auth_value()
        super(GoogleFHIRAuth, self).__init__(state=state)

    @property
    def ready(self):
        """Return True if access_token exists."""
        return True if self.access_token else False

    def reset(self):
        """Clear access_token."""
        super(GoogleFHIRAuth, self).reset()
        self.access_token = None

    def can_sign_headers(self):
        """Return True if access_token exists."""
        return True if self.access_token is not None else False

    def signed_headers(self, headers):
        """Return updated HTTP request headers, if possible, raises if there is no access_token."""
        if not self.can_sign_headers():
            raise Exception("Cannot sign headers since I have no access token")

        if headers is None:
            headers = {}
        headers['Authorization'] = "Bearer {0}".format(self.access_token)

        return headers

    def reauthorize(self, server):
        """Perform reauthorization.

        Args:
            - server - The Server instance to use

        Returns:
            - output - The launch context dictionary, or None on failure
        """
        logger.debug("SMART AUTH: Refreshing token")
        self.access_token = self._get_auth_value()
        return {'access_token': self._access_token}

    def handle_401(self, response, **kwargs):
        """Handle failed requests when authorization failed.

        This gets called after a failed request when an HTTP 401 error
        occurs. This then tries to refresh the access token in the event
        that it expired.

        Args:
            request (object): The failed request object

        """
        if not response.status_code == 401 and not response.status_code == 403:
            return response

        # Free the original connection
        response.content
        response.close()

        # copy the request to resend
        newreq = response.request.copy()

        self.access_token = None
        logger.debug("handle_401, cleared _access_token, retrying with new token")

        self.access_token = self._get_auth_value()
        newreq.headers["Authorization"] = "Bearer {0}".format(self.access_token)

        _response = response.connection.send(newreq, **kwargs)
        _response.history.append(response)
        _response.request = newreq

        return _response

    @property
    def state(self):
        """Save state."""
        s = super(GoogleFHIRAuth, self).state
        if self.access_token is not None:
            s['access_token'] = self.access_token

        return s

    def from_state(self, state):
        """Update ivars from given state information."""
        super(GoogleFHIRAuth, self).from_state(state)
        self.access_token = state.get('access_token') or self.access_token

    def _get_auth_value(self):
        """Return the Authorization header value for the request.

        This gets called when added the Authorization header to the request.
        This fetches the access token from the refresh token if the access token is missing.

        """
        if not self.access_token:
            try:
                # get the local access token using gcloud
                cmd = ['gcloud', 'auth', 'print-access-token']
                logger.debug(f"getting gcloud_access_token {cmd}")
                p = Popen(cmd, stdout=PIPE, stderr=PIPE)
                gcloud_access_token, stderr = p.communicate()
                # remove CR
                gcloud_access_token = gcloud_access_token.decode("utf-8").rstrip()
                assert len(gcloud_access_token) > 0, f'get gcloud_access_token MUST have an access token {stderr}'
                logger.debug(f"gcloud_access_token {gcloud_access_token}")
                self.access_token = gcloud_access_token
                if logger.level == logging.DEBUG:
                    logger.debug(self.access_token)
                    print(self.access_token)
                    # expires_at = datetime.fromisoformat(self.access_token['expires_at'])
                    # logger.debug(f'access token expires in {str(expires_at - now)}')
                    # # add padding
                    # logger.debug(base64.b64decode(self._access_token.split('.')[1] + "==="))
            except Exception as e:
                raise FHIRAuthError(f"Failed to get gcloud_access_token {e}")
        return self.access_token


# register class
GoogleFHIRAuth.register()



"""A setuptools based setup module.

See:
https://packaging.python.org/guides/distributing-packages-using-setuptools/
https://github.com/pypa/anvil
"""


import setuptools.command.build_py
# Always prefer setuptools over distutils
from setuptools import setup, find_packages
from os import path
import subprocess


class DownloadDataIngestionSpreadsheetCommand(setuptools.command.build_py.build_py):
    """Custom build command."""

    description = 'download data from AnVIL data ingestion spreadsheet'

    def run(self):
        """Run the command."""
        self.announce('running data_ingestion_tracker')
        process = subprocess.run(['bin/data_ingestion_tracker'],
                                 stdout=subprocess.PIPE,
                                 universal_newlines=True)
        assert process.returncode == 0, "data_ingestion_tracker must execute successfully"


here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# Arguments marked as "Required" below must be included for upload to PyPI.
# Fields marked as "Optional" may be commented out.

setup(
    # This is the name of your project. The first time you publish this
    # package, this name will be registered for you. It will determine how
    # users can install this project, e.g.:
    #
    # $ pip install pyAnVIL
    #
    # And where it will live on PyPI: https://pypi.org/project/pyAnVIL/
    #
    # There are some restrictions on what makes a valid project name
    # specification here:
    # https://packaging.python.org/specifications/core-metadata/#name
    name='pyAnVIL',  # Required

    # Versions should comply with PEP 440:
    # https://packaging.python.org/en/latest/single_source_version.html
    version='0.0.13rc12',  # Required

    # This is a one-line description or tagline of what your project does. This
    # corresponds to the "Summary" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#summary
    description='AnVIL client library. Combines gen3, terra client APIs with single signon and data harmonization use cases.',  # Optional

    # This field corresponds to the "Description" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#description-optional
    long_description=long_description,  # Optional

    # Denotes that our long_description is in Markdown
    long_description_content_type='text/markdown',  # Optional (see note above)

    # Project's main homepage.
    url='https://github.com/anvilproject/client-apis',  # Optional

    # This should be your name or the name of the organization which owns the
    # project.
    author='The AnVIL project',  # Optional

    # This should be a valid email address corresponding to the author listed
    # above.
    author_email='walsbr@ohsu.edu',  # Optional

    # cmdclass={
    #     'data_ingestion_tracker': DownloadDataIngestionSpreadsheetCommand,
    # },


    # For a list of valid classifiers, see https://pypi.org/classifiers/
    classifiers=[  # Optional
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',

        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here.
        'Programming Language :: Python :: 3.7',
    ],

    # This field adds keywords for your project which will appear on the
    # project page. What does your project relate to?
    #
    # Note that this is a string of words separated by whitespace, not a list.
    keywords='AnVIL terra gen3 bioinformatics',  # Optional

    # You can just specify package directories manually here if your project is
    # simple. Or you can use find_packages().
    #
    # Alternatively, if you just want to distribute a single Python file, use
    # the `py_modules` argument instead as follows, which will expect a file
    # called `my_module.py` to exist:
    #
    #   py_modules=["my_module"],
    #
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),  # Required

    # Specify which Python versions you support. In contrast to the
    # 'Programming Language' classifiers above, 'pip install' will check this
    # and refuse to install the project if the version does not match. If you
    # do not support Python 2, you can simplify this to '>=3.5' or similar, see
    # https://packaging.python.org/guides/distributing-packages-using-setuptools/#python-requires
    python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, <4',

    # This field lists other packages that your project depends on to run.
    # Any package you put here will be installed by pip when your project is
    # installed, so they must be valid existing projects.
    #
    # For an analysis of "install_requires" vs pip's requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=[
        'gen3==4.2.0',
        'firecloud==0.16.29',
        'xmltodict==0.12.0',
        'Click==7.1.2',
        'attrdict==2.0.1',
        'google-cloud-storage==2.0.0',
        'google-cloud-core==2.0.0',
        'fastavro==1.2.0',
        'tabulate==0.8.9',
        'gspread==5.2.0',
        'oauth2client==4.1.3',
        'click_loglevel==0.4.0.post1'
    ],

    # List additional groups of dependencies here (e.g. development
    # dependencies). Users will be able to install these using the "extras"
    # syntax, for example:
    #
    #   $ pip install anvil[dev]
    #
    # Similar to `install_requires` above, these must be valid existing
    # projects.
    # extras_require={  # Optional
    #     'dev': ['check-manifest'],
    #     'test': ['coverage'],
    # },

    # If there are data files included in your packages that need to be
    # installed, specify them here.
    #
    # If using Python 2.6 or earlier, then these have to be included in
    # MANIFEST.in as well.
    package_data={  # Optional
        '': ['*.yaml'],
    },

    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files
    #
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    # data_files=[('my_data', ['data/data_file'])],  # Optional

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # `pip` to create the appropriate form of executable for the target
    # platform.
    #
    # For example, the following would provide a command called `sample` which
    # executes the function `main` from this package when invoked:
    # entry_points={  # Optional
    #     'console_scripts': [
    #         'sample=sample:main',
    #     ],
    # },

    # TODO - not working
    # entry_points={
    #     'console_scripts': [
    #         'data_ingestion_tracker=anvil.bin.data_ingestion_tracker:data_ingestion_tracker',
    #         'reconciler=anvil.bin.reconciler:reconciler',
    #     ],
    # },
    scripts="""
        bin/anvil_etl
        bin/fhir_curl
        bin/fhir_env
        bin/fhir_setup
        bin/fhir_load        
    """.split(),
    include_package_data=True,

    # List additional URLs that are relevant to your project as a dict.
    #
    # This field corresponds to the "Project-URL" metadata fields:
    # https://packaging.python.org/specifications/core-metadata/#project-url-multiple-use
    #
    # Examples listed include a pattern for specifying where the package tracks
    # issues, where the source is hosted, where to say thanks to the package
    # maintainers, and where to support the project financially. The key is
    # what's used to render the link text on PyPI.
    project_urls={  # Optional
        'Bug Reports': 'https://github.com/anvilproject/client-apis/issues',
        'Source': 'https://github.com/anvilproject/client-apis/',
    },
)



# set up a command library
import logging
from collections import defaultdict
from urllib.parse import urlparse

from attrdict import AttrDict

from anvil.etl.transformers import LogCapture
from anvil.etl.utilities.entities import Entities

# logger = LogCapture('anvil.etl.transformers.normalizer')
# logger.addHandler(logging.root.handlers[0])

# from contextvars import ContextVar
# logger = ContextVar('logger').get()
from . import logger


#
# runner
#
from ..extractors.gen3 import DRSReader


def exec_command(context, consortium_name, workspace, config, cmd):
    """Find and run version of the command that matches."""
    workspace_name = workspace['workspace']['name']

    paths = [
        f"{consortium_name}/{workspace_name}/{cmd}",
        f"{consortium_name}//{cmd}",
        f"//{workspace_name}/{cmd}",
        f"//{cmd}"
    ]

    path = None
    for path in paths:
        try:
            runner = C.get(path, None)
            if runner:
                return runner(context, consortium_name, workspace, config)
        except Exception as e:
            raise e

    assert False, f"Misconfiguration, we shouldn't get here. {cmd} {path} {consortium_name} {workspace_name}"


def _extract_bucket_fields(output_path=None, entities=None, workspace_name=None):
    """Query db, sample first row in all entities, determine fields that have references to bucket object."""
    if not entities:
        assert output_path
        entities = Entities(path=f"{output_path}/terra_entities.sqlite")
    for workspace in entities.get_by_label('workspace'):
        workspace = AttrDict(workspace)
        if workspace_name and workspace.workspace.name != workspace_name:
            continue
        schema = AttrDict(entities.get_edges_by_label(src=workspace.workspace.name, src_name='workspace', dst_name='schema'))
        for entity_name in schema.schema.keys():
            if entity_name == 'schema':
                continue
            child = entities.get_edges_by_label(src=workspace.workspace.name, src_name='workspace', dst_name=entity_name, limit=1)
            if entity_name not in child:
                logger.error(('invalid.child', workspace.consortium_name, workspace.workspace.name, entity_name, child))
                continue
            bucket_keys = [k for k, v in child[entity_name]['attributes'].items() if isinstance(v, str) and v.startswith('gs://')]
            if len(bucket_keys) == 0:
                continue
            buckets = [child[entity_name]['attributes'][key] for key in bucket_keys]
            buckets = list(set([urlparse(b).netloc for b in buckets]))
            # TODO - move this exception to config or other ?
            if entity_name == 'sequencing' and 'AnVIL_CMG_Broad' in workspace.workspace.name:
                bucket_keys = 'crai_or_bai_path,crai_path,cram_or_bam_path,cram_path,md5_path'.split(',')
            yield {
                'consortium_name': workspace.consortium_name,
                'workspace_name': workspace.workspace.name,
                'entity_name': entity_name,
                'bucket_fields': bucket_keys,
                'buckets': buckets
            }


def _recursive_default_dict():
    """Recursive default dict, any key defaults to a dict."""
    return defaultdict(_recursive_default_dict)


def consortium_config(context, consortium_name, workspace, config):
    context.consortium_config = config[consortium_name]
    inverted_config = _recursive_default_dict()
    for fhir_entity in context.consortium_config['entities']:
        for alias in context.consortium_config['entities'][fhir_entity]['aliases']:
            inverted_config[alias] = fhir_entity
    context.inverted_config = inverted_config


def patient_entity_names(context, consortium_name, workspace, config):
    context.patient_entity_names = context.consortium_config['entities']['Patient']['aliases']


def patient_entity_names_CMG(context, consortium_name, workspace, config):
    patient_entity_names(context, consortium_name, workspace, config)


def patient_entity_names_CCDG(context, consortium_name, workspace, config):
    patient_entity_names(context, consortium_name, workspace, config)


def specimen_entity_name(context, consortium_name, workspace, config):
    entities_with_bucket_fields = [bf['entity_name'] for bf in context['bucket_fields']]
    specimen_aliases_with_buckets = []
    for alias in context.consortium_config['entities']['Specimen']['aliases']:
        if alias in entities_with_bucket_fields:
            specimen_aliases_with_buckets.append(alias)

    if len(specimen_aliases_with_buckets) == 0:
        context.specimen_entity_name = context.consortium_config['entities']['Specimen']['aliases'][0]
        logger.warning(('specimen.entity.has.no.bucket.fields', workspace.workspace.name, context.specimen_entity_name))
    elif 'sample' in specimen_aliases_with_buckets:
        context.specimen_entity_name = 'sample'
    else:
        context.specimen_entity_name = specimen_aliases_with_buckets[0]


def patient_from_specimen_numbered(context, consortium_name, workspace, config):
    """Look for workspace.specimen.01-subject_id -> workspace.subject."""

    return context.specimen['attributes']['01-subject_id']


def patient_from_specimen_subject_id(context, consortium_name, workspace, config):
    """Look for workspace.specimen.subject_id -> workspace.subject."""

    return context.specimen['attributes']['subject_id']


def patient_from_specimen_participant_id(context, consortium_name, workspace, config):
    """Look for context.specimen.participant_id -> workspace.subject."""

    return context.specimen['attributes']['participant_id']


def patient_from_specimen_subject(context, consortium_name, workspace, config):
    """Look for context.specimen.subject -> workspace.subject."""

    return context.specimen['attributes']['subject']


def patient_from_participant(context, consortium_name, workspace, config):
    """Look for context.specimen.subject -> workspace.subject."""

    return context.specimen['attributes']['participant']


def patient_from_specimen(context, consortium_name, workspace, config):
    """Introspect specimen, see if it has a dict pointing to subject."""

    patient_entity_name = None
    for n in context.patient_entity_names:
        if n in context.specimen['attributes']:
            patient_entity_name = n
            break

    if not patient_entity_name:
        if 'no.find.patient_entity_name' not in context['logged_already']:
            logger.error(('no.find.patient_entity_name', consortium_name, workspace.workspace.name,
                          context.patient_entity_names, list(workspace.children.keys())))  # list since cannot pickle 'dict_keys' object
            context['logged_already'].append('no.find.patient_entity_name')
        return None

    if patient_entity_name not in context.specimen['attributes']:
        if 'specimen.missing.patient' not in context['logged_already']:
            logger.error(('specimen.missing.patient', consortium_name, workspace.workspace.name, context.specimen))
            context['logged_already'].append('specimen.missing.patient')
        return None

    patient_key_from_specimen = context.specimen['attributes'][patient_entity_name]
    if isinstance(patient_key_from_specimen, dict):
        patient_key_from_specimen = patient_key_from_specimen['entityName']
    return patient_key_from_specimen


def patient_from_participant_pfb_subject(context, consortium_name, workspace, config):
    """Introspect specimen, see if it has a dict pointing to subject."""

    patient_entity_name = None
    for n in context.patient_entity_names:
        if n in context.specimen['attributes']:
            patient_entity_name = n
            break

    if not patient_entity_name:
        if 'no.find.patient_entity_name' not in context['logged_already']:
            logger.error(('no.find.patient_entity_name', consortium_name, workspace.workspace.name))
            context['logged_already'].append('no.find.patient_entity_name')
        return None

    if patient_entity_name not in context.specimen['attributes']:
        if 'specimen.missing.patient' not in context['logged_already']:
            logger.error(('specimen.missing.patient', consortium_name, workspace.workspace.name, context.specimen))
            context['logged_already'].append('specimen.missing.patient')
        return None

    patient_key_from_specimen = context.specimen['attributes'][patient_entity_name]
    if isinstance(patient_key_from_specimen, dict):
        patient_key_from_specimen = f"pfb:{patient_key_from_specimen['entityName']}"
    return patient_key_from_specimen


def patient_from_qc_result_sample(context, consortium_name, workspace, config):
    """Introspect specimen, see if it has a dict pointing to subject."""
    sample_id = context.specimen['attributes']['qc_result_sample']
    for s in workspace.children['sample']:
        if s['name'] == sample_id or s['attributes'].get('collaborator_sample_id', None) == sample_id:
            for k in ['subject_id', 'participant_id', 'collaborator_participant_id']:
                if k in s['attributes']:
                    return s['attributes'][k]
            logger.error(('no.subject.found', workspace.workspace.name, sample_id, s))


def patient_from_specimen_all(context, consortium_name, workspace, config):
    """Try all commands."""
    cmd = None
    if 'subject' in context.specimen['attributes']:
        cmd = patient_from_specimen_subject
    if 'participant_id' in context.specimen['attributes']:
        cmd = patient_from_specimen_participant_id
    if '01-subject_id' in context.specimen['attributes']:
        cmd = patient_from_specimen_numbered
    if 'subject_id' in context.specimen['attributes']:
        cmd = patient_from_specimen_subject_id
    if 'specimen' in context.specimen['attributes']:
        cmd = patient_from_specimen
    if 'participant' in context.specimen['attributes']:
        cmd = patient_from_participant
    if 'pfb:subject' in context.specimen['attributes']:
        cmd = patient_from_participant_pfb_subject
    if 'qc_result_sample' in context.specimen['attributes']:
        cmd = patient_from_qc_result_sample

    if cmd:
        return cmd(context, consortium_name, workspace, config)
    else:
        if 'unparsable.specimen' not in context['logged_already']:
            logger.error(('unparsable.specimen', consortium_name, workspace.workspace.name, context.specimen))
            context['logged_already'].append('unparsable.specimen')
        return None


def task_entity_names(context, consortium_name, workspace, config):
    """Set task_entity_names on context to aliases, or _implied if no Task configured."""
    if 'Task' in context.consortium_config['entities']:
        context.task_entity_names = context.consortium_config['entities']['Task']['aliases']
    else:
        context.task_entity_names = ['_implied']


def ensure_tasks(context, consortium_name, workspace, config):
    """Set tasks on workspace to a workspace child, or create an artificial one if _implied."""
    workspace['tasks'] = {}
    if '_implied' in context.task_entity_names:
        workspace['tasks']['_implied'] = _recursive_default_dict()
    else:
        for task_entity_name in context.task_entity_names:
            if task_entity_name in workspace.children.keys():
                workspace['tasks'][task_entity_name] = workspace.children[task_entity_name]
        if not workspace['tasks']:
            workspace['tasks']['_implied'] = _recursive_default_dict()


def ensure_bucket_fields(context, consortium_name, workspace, config):
    """Introspect workspace schema looking for fields that have bucket objects."""

    context['bucket_fields'] = [bf for bf in
                                _extract_bucket_fields(entities=context['entities'], workspace_name=workspace.workspace.name)]


def ensure_bucket_fieldsAnVIL_CMG_BaylorHopkins_HMB_NPU_WES(context, consortium_name, workspace, config):
    """Override sample.bucket_fields"""
    ensure_bucket_fields(context, consortium_name, workspace, config)
    assert len(context['bucket_fields']) == 1
    context['bucket_fields'][0]['bucket_fields'] = "bam,bam_md5,crai,cram,cram_md5".split(',')


def ensure_bucket_fieldsAnVIL_CMG_Broad_crai_or_bai(context, consortium_name, workspace, config):
    """Override sample.bucket_fields"""
    ensure_bucket_fields(context, consortium_name, workspace, config)
    assert len(context['bucket_fields']) == 1
    context['bucket_fields'][0][
        'bucket_fields'] = 'crai_or_bai_path,crai_path,cram_or_bam_path,cram_path,md5_path'.split(',')


def ensure_tasks_populated(context, consortium_name, workspace, config):
    """Create a terra style entity for an implied task if none exists."""
    if '_implied' in workspace['tasks']:
        assert workspace.specimens
        workspace['tasks']['_implied'] = []
        for specimen in workspace.specimens:
            # create terra type entity
            workspace['tasks']['_implied'].append(AttrDict({
                'entityType': 'Task',
                'name': specimen['name'],
                'attributes': {
                    # TODO - are there any imputed attributes?
                }
            }))

    # check that other keys exist
    assert len(workspace['tasks']) > 0, ('missing.tasks', consortium_name, workspace.workspace.name)


def extract_specimen_reference_sample_alias(context, consortium_name, workspace, config):
    """Extract a specimen key from arbitrary entity, reads from ctx.entity_with_specimen_reference, sets specimen_reference aka the key."""
    assert 'entity_with_specimen_reference' in context
    entity = context['entity_with_specimen_reference']
    if 'sample_alias' in entity['attributes']:
        context['specimen_reference'] = {
            'name': entity['attributes']['sample_alias'],
            'entityType': 'sample',
            'fhir_entity': context.inverted_config['sample']
        }


def extract_specimen_reference_sample_id(context, consortium_name, workspace, config):
    """Extract a specimen key from arbitrary entity, reads from ctx.entity_with_specimen_reference, sets specimen_reference aka the key."""
    assert 'entity_with_specimen_reference' in context
    entity = context['entity_with_specimen_reference']
    if 'sample_id' in entity['attributes']:
        context['specimen_reference'] = {
            'name': entity['attributes']['sample_id'],
            'entityType': 'sample',
            'fhir_entity': context.inverted_config['sample']
        }


def extract_specimen_reference_collaborator_sample_id(context, consortium_name, workspace, config):
    """Extract a specimen key from arbitrary entity, reads from ctx.entity_with_specimen_reference, sets specimen_reference aka the key."""
    assert 'entity_with_specimen_reference' in context
    entity = context['entity_with_specimen_reference']
    if 'collaborator_sample_id' in entity['attributes']:
        context['specimen_reference'] = {
            'name': entity['attributes']['collaborator_sample_id'],
            'entityType': 'sample',
            'fhir_entity': context.inverted_config['sample']
        }


def extract_specimen_reference_CMG(context, consortium_name, workspace, config):
    """Extract a specimen key from arbitrary entity, reads from ctx.entity_with_specimen_reference, sets specimen_reference aka the key."""
    assert 'entity_with_specimen_reference' in context
    entity = context['entity_with_specimen_reference']
    for key in ['collaborator_sample_id', 'sample', 'sample_alias']:
        if key in entity['attributes']:
            context['specimen_reference'] = {
                'name': entity['attributes'][key],
                'entityType': 'sample',
                'fhir_entity': context.inverted_config['sample']
            }


def ensure_tasks_linked_to_documents(context, consortium_name, workspace, config):
    """Ensure tasks have input that is a specimen, and outputs that are document references."""

    # set up a hash of children with buckets for quick lookup
    _children = _recursive_default_dict()
    for bf in context.bucket_fields:
        if bf['entity_name'] in workspace.children:
            for child in workspace.children[bf['entity_name']]:
                _children[bf['entity_name']][child['name']] = child
    # create a task, whose inputs are entities with bucket fields.
    # and whose outputs are those bucket fields
    deletion_keys = []
    for task_entity in workspace['tasks']:
        document_count = 0
        for task in workspace['tasks'][task_entity]:
            _task_children = _recursive_default_dict()
            _task_inputs = []
            for bf in context.bucket_fields:
                if bf['entity_name'] in _children:
                    if task['name'] in _children[bf['entity_name']]:
                        child = _children[bf['entity_name']][task['name']]
                        _task_inputs.append(
                            {
                                'entityType': child['entityType'],
                                'name': child['name'],
                                'fhir_entity': context.inverted_config[child['entityType']]
                            }
                        )
                        for field in bf['bucket_fields']:
                            if workspace.workspace.name not in ['AnVIL_CMG_BaylorHopkins_HMB-NPU_WES']:
                                # TODO - remove if not needed
                                if 'AnVIL_CMG_Broad' not in workspace.workspace.name:
                                    assert field in child['attributes'], (field, child, bf)
                            if field in child['attributes']:
                                _task_children[bf['entity_name']][field] = child['attributes'][field]
                                document_count += 1
                # # a task might have it's own bucket fields
                # for field in bf['bucket_fields']:
                #     logger.error(('?', field))
                #     if field in task['attributes']:
                #         logger.error(('?', bf['entity_name'], field))
                #         _task_children[bf['entity_name']][field] = task['attributes'][field]
                #         document_count += 1

            if 'Specimen' not in [ti['fhir_entity'] for ti in _task_inputs]:
                no_specimen = True
                for ti in _task_inputs:
                    context.entity_with_specimen_reference = _children[ti['entityType']][ti['name']]
                    exec_command(context, consortium_name, workspace, config, 'extract_specimen_reference')
                    if 'specimen_reference' in context:
                        _task_inputs.append(context['specimen_reference'])
                        no_specimen = False
                        del context['specimen_reference']
                if no_specimen and len(_task_inputs) > 0:
                    logger.error(('no.specimen.in.task', consortium_name, workspace.workspace.name, _task_inputs))

            # It is possible for an entity that has  bucket fields to be classified as a task,
            # make sure we don't self reference
            _task_inputs = [ti for ti in _task_inputs if
                            not (task['name'] == ti['name'] and task['entityType'] == ti['entityType'])]

            task['outputs'] = _task_children
            task['inputs'] = _task_inputs
            if document_count == 0:
                deletion_keys.append(task_entity)

    # delete task entities without documents
    for key in set(deletion_keys):
        del workspace['tasks'][key]
        logger.warning(('no.task', consortium_name, workspace.workspace.name, key))


def blob_attributes(context, consortium_name, workspace, config):
    """Decorate task outputs with blob attributes."""
    google_entities = None
    if context['validate_buckets']:
        google_entities = Entities(path=f"{context['output_path']}/google_entities.sqlite")
    drs_reader = DRSReader(context['output_path'])

    missing_blob_count = 0
    for task_entity in workspace['tasks']:
        for task in workspace['tasks'][task_entity]:
            for output_source, output in task['outputs'].items():
                for output_property, url in output.items():
                    if context['validate_buckets']:
                        blob = google_entities.get(url)
                        if not blob:
                            blob = {'url': url, 'drs_uri': None}
                            missing_blob_count += 1
                            if 'blob.not.in.bucket' not in context['logged_already']:
                                logger.warning(('blob.not.in.bucket', consortium_name, workspace.workspace.name,
                                                output_source, output_property, url))
                                context['logged_already'].append('blob.not.in.bucket')
                    else:
                        blob = {'url': url, 'drs_uri': None}
                    drs = drs_reader.get(url.split('/')[-1])
                    if drs:
                        blob['drs_uri'] = drs.get('ga4gh_drs_uri', None)
                    task['outputs'][output_source][output_property] = blob
    if missing_blob_count > 0:
        logger.error(('total.blob.not.in.bucket', consortium_name, workspace.workspace.name, missing_blob_count))


def practitioner(context, consortium_name, workspace, config):
    """Determine the PI, set practitioner on workspace."""
    if 'study_pi' in workspace.workspace.attributes:
        workspace.practitioner = workspace.workspace.attributes['study_pi']
    elif 'library:institute' in workspace.workspace.attributes:
        workspace.practitioner = workspace.workspace.attributes['library:institute']['items'][0]
    else:
        logger.error(
            ('no.find.practitioner', consortium_name, workspace.workspace.name, workspace.workspace.attributes))
        workspace.practitioner = 'NA'
    if not workspace.practitioner or len(workspace.practitioner) == 0:
        workspace.practitioner = 'NA'


def phenotypes(context, consortium_name, workspace, config):
    """Retrieve phenotypes from patient."""
    context['patient']['phenotype_values'] = []


def phenotypes_fields_CMG(context, consortium_name, workspace, config):
    """Retrieve phenotypes from patient."""
    if 'phenotype_fields' not in context:
        schema = workspace.children['schema'][0]
        phenotype_fields = []
        for entity_name in schema.keys():
            for attribute_name in schema[entity_name]['attributeNames']:
                for k in ['pheno', 'hpo', 'genotype']:
                    if k in attribute_name:
                        phenotype_fields.append((entity_name, attribute_name))
        context['phenotype_fields'] = phenotype_fields

    phenotype_values = []
    for key, value in context['patient']['attributes'].items():
        for entity_name, attribute_name in context['phenotype_fields']:
            if key == attribute_name:
                phenotype_values.append((key, value))
    context['patient']['phenotype_values'] = phenotype_values


def phenotypes_CMG(context, consortium_name, workspace, config):
    """Retrieve phenotypes from patient."""
    phenotypes_fields_CMG(context, consortium_name, workspace, config)
    phenotype_values = defaultdict(set)
    for attribute_name, value in context['patient']['phenotype_values']:
        if not value or value == '-':
            continue
        for k in ['hpo_present']:
            if k in attribute_name:
                phenotype_values['present'].update(value.split('|'))
        for k in ['phenotype_group']:
            if k in attribute_name:
                phenotype_values['group'].update(value.split('|'))
        for k in ['hpo_absent']:
            if k in attribute_name:
                phenotype_values['absent'].update(value.split('|'))
        for k in ['phenotype_description']:
            if k in attribute_name:
                phenotype_values['description'].update(value.split('|'))
    # xform to a list
    for k, v in phenotype_values.items():
        phenotype_values[k] = list(v)
    context['patient']['phenotype_values'] = [phenotype_values]


def phenotypes_CCDG(context, consortium_name, workspace, config):
    """Retrieve phenotypes from patient."""
    phenotypes(context, consortium_name, workspace, config)
    if len(context['patient']['phenotype_values']) > 0:
        logger.error(('TODO.phenotypes_CCDG', context['patient']['phenotype_values']))
    context['patient']['phenotype_values'] = []


def disease(context, consortium_name, workspace, config):
    """Retrieve disease from patient."""
    context['patient']['disease_values'] = []


def disease_CMG(context, consortium_name, workspace, config):
    """Retrieve disease from patient."""
    disease_values = defaultdict(set)
    for match in ['disease', 'affected']:
        for key, value in context['patient']['attributes'].items():
            if match in key:
                if value.lower() in ["n/a", 'na', 'nan', '-', '']:
                    continue
                values = [v for v in value.split('|') if len(v) > 0]
                disease_values[key].update(values)

    if 'affected_status' in disease_values and 'disease_description' in disease_values:
        disease_values['description'].update(
            disease_values['disease_description'])  # = list(disease_values['disease_description'])[0],
        disease_values['present'].update(
            disease_values['affected_status'])  # = list(disease_values['affected_status'])[0]

    context['patient']['disease_values'] = [disease_values]


def disease_NIMH(context, consortium_name, workspace, config):
    """Retrieve disease from patient."""
    disease_CMG(context, consortium_name, workspace, config)


def disease_CCDG(context, consortium_name, workspace, config):
    """Retrieve disease from patient."""
    disease_values = defaultdict(set)
    for attribute_name in ['ADI-R_DIAG', 'affected_status', 'affected_status_detail', 'affection_status',
                           'atrial_fibrillation', 'atrial_fibrilliation', 'crohns_disease', 'disease_description',
                           'epilepsy', 'hemorrhagic_stroke', 'ischemic_stroke', 'ulcerative_colitis']:
        for key, value in context['patient']['attributes'].items():
            if key == attribute_name:
                if value.lower() in ["n/a", 'na', 'nan']:
                    continue
                disease_values[key].add(value)

    if 'affected_status' in disease_values and 'disease_description' in disease_values:
        disease_values['description'].update(
            disease_values['disease_description'])  # = list(disease_values['disease_description'])[0],
        disease_values['present'].update(
            disease_values['affected_status'])  # = list(disease_values['affected_status'])[0]
    context['patient']['disease_values'] = [disease_values]


def disease_CCDG_ADI_R_DIAG(context, consortium_name, workspace, config):
    """Retrieve disease from patient. AnVIL_CCDG_NYGC_NP_Autism_ACE2_GRU-MDS_WGS"""
    disease_CCDG(context, consortium_name, workspace, config)
    disease_values_list = []

    disease_values = context['patient']['disease_values']
    if "ADI-R_DIAG" in disease_values:
        description = disease_values["ADI-R_DIAG"]
        present = disease_values['affection_status']
        disease_values_list.append({
            'description': description,
            'present': present
        })

    for k in ["atrial_fibrilliation", "crohns_disease", "epilepsy", "hemorrhagic_stroke", "ulcerative_colitis"]:
        if k in disease_values:
            disease_values_list.append({
                'description': k.replace('_', ' '),
                'present': disease_values[k]
            })

    context['patient']['disease_values'] = disease_values_list


def gender(context, consortium_name, workspace, config):
    """Retrieve gender from patient."""
    context['patient']['gender'] = None


def gender_NHGRI(context, consortium_name, workspace, config):
    """Retrieve gender from patient."""
    assert 'patient' in context
    keys = ["gender"]
    for k in keys:
        if k in context['patient']['attributes']:
            context['patient']['gender'] = context['patient']['attributes'][k].lower()
            assert context['patient']['gender'] in "male,female,other,unknown".split(',')


def gender_NIMH(context, consortium_name, workspace, config):
    """Retrieve gender from patient."""
    gender_CMG(context, consortium_name, workspace, config)


def gender_CMG(context, consortium_name, workspace, config):
    """Retrieve gender from patient."""
    assert 'patient' in context
    keys = ["sex"]
    context['patient']['gender'] = None
    for k in keys:
        if k in context['patient']['attributes']:
            value = context['patient']['attributes'][k].lower()
            if value in ['-', 'not reported', 'notreported', 'na', 'unspecified']:
                value = 'unknown'
            if value not in "male,female,other,unknown".split(','):
                logger.error(('unrecognized.gender', value))
                value = 'unknown'
            context['patient']['gender'] = value
            assert context['patient']['gender'] in "male,female,other,unknown".split(','), context['patient']['gender']


def gender_CCDG(context, consortium_name, workspace, config):
    """Retrieve gender from patient."""
    gender_CMG(context, consortium_name, workspace, config)


def family_relationship(context, consortium_name, workspace, config):
    """Retrieve family relationship from patient."""
    family_relationship_values = {}
    for match in ['family_id', 'proband_relationship', 'family_relationship', 'maternal', 'paternal', 'father',
                  'mother']:
        for key, value in context['patient']['attributes'].items():
            if match in key.lower():
                if value.lower() in ["#n/a", "n/a", 'na', 'nan', '-', '']:
                    continue
                values = [v for v in value.split('|') if len(v) > 0]
                assert len(values) == 1, values
                family_relationship_values[match] = values[0]

    context['patient']['family_relationship_values'] = [family_relationship_values]

    if len(family_relationship_values) > 0:
        assert 'FamilyRelationship' in context.consortium_config['entities']
        if 'FamilyRelationship' in context.consortium_config['entities']:
            for alias in context.consortium_config['entities']['FamilyRelationship']['aliases']:
                if alias in workspace['children']:
                    for family in workspace['children'][alias]:
                        if family['name'] == family_relationship_values['family_id']:
                            context['patient']['family'] = family
            if 'family' not in context['patient']:
                logger.error(('no.family', context['patient']))

# def family_relationship_CMG(context, consortium_name, workspace, config):
#     """Retrieve family relationship from patient."""
#     ["family_id", "family_relationship", "maternal_id", "paternal_id"]
#     pass

def body_site(context, consortium_name, workspace, config):
    """Retrieve body site from specimen."""
    pass


def body_site_CMG(context, consortium_name, workspace, config):
    """Retrieve body site from specimen."""
    # TODO ["sample_source"]  
    pass


def link_specimen_to_patient(context, consortium_name, workspace, config):
    """Standard model specimen->patient."""

    # start traversal from specimen
    specimens = workspace.children.get(context.specimen_entity_name, None)
    if not specimens or len(specimens) == 0:
        logger.error(('no.find.specimens', consortium_name, workspace.workspace.name))
        return
    workspace.specimens = specimens

    # traverse up to patient
    logged_already = False
    specimen_patient_ids = []
    missing_patient_count = 0
    for specimen in workspace.specimens:
        context.specimen = specimen
        patient_id = exec_command(context, consortium_name, workspace, config, 'patient_from_specimen')
        if not patient_id:
            if not logged_already:
                logged_already = True
                logger.error((
                             "no.find.patient.from.sample", consortium_name, workspace.workspace.name, context.specimen,
                             list(workspace.children.keys())))
            missing_patient_count += 1
            continue
        else:
            specimen_patient_ids.append((specimen.name, patient_id))
    if missing_patient_count:
        logger.error(
            ("total.no.find.patient.from.sample", consortium_name, workspace.workspace.name, missing_patient_count))
    # store sample->patient on workspace for now
    total_patient_count = 0
    for patient_entity_name in context.patient_entity_names:
        # patient_entity_names are possibilities, they don't have to exist
        total_patient_count += len(workspace.children.get(patient_entity_name, []))
    workspace.specimen_patient_ids = specimen_patient_ids
    workspace.total_patient_count = total_patient_count


def patient_model(context, consortium_name, workspace, config):
    """Standard model patient->specimen->task->document_reference."""
    patients = {}
    specimens = {}
    if 'patient_entity_names' not in context:
        logger.warning(('missing.context.patient_entity_names', consortium_name, workspace.workspace.name))
        return
    for patient_entity_name in context.patient_entity_names:
        if patient_entity_name in workspace.children:
            for p in workspace.children[patient_entity_name]:
                # normalize patient properties
                context['patient'] = p
                exec_command(context, consortium_name, workspace, config, 'phenotypes')
                exec_command(context, consortium_name, workspace, config, 'disease')
                exec_command(context, consortium_name, workspace, config, 'gender')
                exec_command(context, consortium_name, workspace, config, 'family_relationship')
                # hash by several keys, the name and any other we know about
                patients[p['name']] = p
                for k in ['collaborator_participant_id']:
                    if k in p['attributes']:
                        patients[p['attributes'][k]] = p

    if len(patients) == 0:
        logger.error(('no.patients.found', consortium_name, workspace.workspace.name,
                      f"{context.patient_entity_names} not found in {workspace.children.keys()}"))
        return

    assert context.specimen_entity_name in workspace.children.keys()
    for s in workspace.children[context.specimen_entity_name]:
        # normalize specimen properties
        context['specimen'] = s
        exec_command(context, consortium_name, workspace, config, 'body_site')
        # hash by several keys, the name and any other we know about
        specimens[s['name']] = s
        for k in ['collaborator_sample_id']:
            if k in s['attributes']:
                specimens[s['attributes'][k]] = s

    for task_entity in workspace.tasks:
        for task in workspace.tasks[task_entity]:
            for input in task['inputs']:
                if input['fhir_entity'] == 'Specimen':
                    if input['name'] not in specimens:
                        logger.error(('no.specimen.found.for.task', consortium_name, workspace.workspace.name,
                                      task['entityType'], input['name']))
                        continue
                    if 'tasks' not in specimens[input['name']]:
                        specimens[input['name']]['tasks'] = []
                    specimens[input['name']]['tasks'].append(task)

    missing_patient_count = 0
    for specimen_id, patient_id in workspace.specimen_patient_ids:
        # assert patient_id, (specimen_id, patient_id)
        if patient_id and not isinstance(patient_id, str):
            # hmmm - should we always just make this a str
            patient_id = patient_id['entityName']
        if patient_id not in patients:
            missing_patient_count += 1
            if "no.find.patient.specified.by.specimen" not in context.logged_already:
                logger.error(("no.find.patient.specified.by.specimen", consortium_name, workspace.workspace.name,
                              context.specimen_entity_name, patient_id))
                context['logged_already'].append("no.find.patient.specified.by.specimen")
            continue
        if specimen_id not in specimens:
            logger.error(("no.find.specimen", consortium_name, workspace.workspace.name, specimen_id))
            continue
        if 'specimens' not in patients[patient_id]:
            patients[patient_id]['specimens'] = []
        patients[patient_id]['specimens'].append(specimens[specimen_id])
    if missing_patient_count:
        logger.error(("total.no.find.patient.specified.by.specimen", consortium_name, workspace.workspace.name,
                      missing_patient_count))
    # set on workspace
    workspace.patients = patients


#
# command specializations hierarchy
#
C = {}
C['//consortium_config'] = lambda context, consortium_name, workspace, config: consortium_config(context, consortium_name, workspace, config)
C['CMG//patient_entity_names'] = lambda context, consortium_name, workspace, config: patient_entity_names_CMG(context, consortium_name, workspace, config)
C['CCDG//patient_entity_names'] = lambda context, consortium_name, workspace, config: patient_entity_names_CCDG(context, consortium_name, workspace, config)
C['//patient_entity_names'] = lambda context, consortium_name, workspace, config: patient_entity_names(context, consortium_name, workspace, config)
C['//specimen_entity_name'] = lambda context, consortium_name, workspace, config: specimen_entity_name(context, consortium_name, workspace, config)
C['//patient_from_specimen'] = lambda context, consortium_name, workspace, config: patient_from_specimen_all(context, consortium_name, workspace, config)
C['//task_entity_names'] = lambda context, consortium_name, workspace, config: task_entity_names(context, consortium_name, workspace, config)
C['//ensure_tasks'] = lambda context, consortium_name, workspace, config: ensure_tasks(context, consortium_name, workspace, config)
C['//ensure_tasks_populated'] = lambda context, consortium_name, workspace, config: ensure_tasks_populated(context, consortium_name, workspace, config)
C['//ensure_tasks_linked_to_documents'] = lambda context, consortium_name, workspace, config: ensure_tasks_linked_to_documents(context, consortium_name, workspace, config)
C['//ensure_bucket_fields'] = lambda context, consortium_name, workspace, config: ensure_bucket_fields(context, consortium_name, workspace, config)
C['CMG/AnVIL_CMG_BaylorHopkins_HMB-NPU_WES/ensure_bucket_fields'] = lambda context, consortium_name, workspace, config: ensure_bucket_fieldsAnVIL_CMG_BaylorHopkins_HMB_NPU_WES(context, consortium_name, workspace, config)
C['CMG/AnVIL_CMG_Broad_Kidney_Hildebrandt_WES/ensure_bucket_fields'] = lambda context, consortium_name, workspace, config: ensure_bucket_fieldsAnVIL_CMG_Broad_crai_or_bai(context, consortium_name, workspace, config)
C['CMG/AnVIL_CMG_Broad_Muscle_KNC_WES/ensure_bucket_fields'] = lambda context, consortium_name, workspace, config: ensure_bucket_fieldsAnVIL_CMG_Broad_crai_or_bai(context, consortium_name, workspace, config)
C['//blob_attributes'] = lambda context, consortium_name, workspace, config: blob_attributes(context, consortium_name, workspace, config)

C['CCDG//extract_specimen_reference'] = lambda context, consortium_name, workspace, config: extract_specimen_reference_sample_id(context, consortium_name, workspace, config)
C['CMG//extract_specimen_reference'] = lambda context, consortium_name, workspace, config: extract_specimen_reference_CMG(context, consortium_name, workspace, config)
C['NIMH//extract_specimen_reference'] = lambda context, consortium_name, workspace, config: extract_specimen_reference_sample_id(context, consortium_name, workspace, config)

C['//practitioner'] = lambda context, consortium_name, workspace, config: practitioner(context, consortium_name, workspace, config)

C['//phenotypes'] = lambda context, consortium_name, workspace, config: phenotypes(context, consortium_name, workspace, config)
C['CMG//phenotypes'] = lambda context, consortium_name, workspace, config: phenotypes_CMG(context, consortium_name, workspace, config)
C['CCDG//phenotypes'] = lambda context, consortium_name, workspace, config: phenotypes_CCDG(context, consortium_name, workspace, config)

C['//disease'] = lambda context, consortium_name, workspace, config: disease(context, consortium_name, workspace, config)
C['CMG//disease'] = lambda context, consortium_name, workspace, config: disease_CMG(context, consortium_name, workspace, config)
C['CCDG//disease'] = lambda context, consortium_name, workspace, config: disease_CCDG(context, consortium_name, workspace, config)
C['NIMH//disease'] = lambda context, consortium_name, workspace, config: disease_NIMH(context, consortium_name, workspace, config)
C['NHGRI//disease'] = lambda context, consortium_name, workspace, config: disease_NIMH(context, consortium_name, workspace, config)

C['CCDG/AnVIL_CCDG_NYGC_NP_Autism_ACE2_GRU-MDS_WGS/disease'] = lambda context, consortium_name, workspace, config: disease_CCDG_ADI_R_DIAG(context, consortium_name, workspace, config)

C['//gender'] = lambda context, consortium_name, workspace, config: gender(context, consortium_name, workspace, config)
C['CMG//gender'] = lambda context, consortium_name, workspace, config: gender_CMG(context, consortium_name, workspace, config)
C['NHGRI//gender'] = lambda context, consortium_name, workspace, config: gender_NHGRI(context, consortium_name, workspace, config)
C['NIMH//gender'] = lambda context, consortium_name, workspace, config: gender_NIMH(context, consortium_name, workspace, config)
C['CCDG//gender'] = lambda context, consortium_name, workspace, config: gender_CCDG(context, consortium_name, workspace, config)

C['//family_relationship'] = lambda context, consortium_name, workspace, config: family_relationship(context, consortium_name, workspace, config)

C['//body_site'] = lambda context, consortium_name, workspace, config: body_site(context, consortium_name, workspace, config)
C['CMG//body_site'] = lambda context, consortium_name, workspace, config: body_site_CMG(context, consortium_name, workspace, config)

C['//link_specimen_to_patient'] = lambda context, consortium_name, workspace, config: link_specimen_to_patient(context, consortium_name, workspace, config)

C['//patient_model'] = lambda context, consortium_name, workspace, config: patient_model(context, consortium_name, workspace, config)



"""Retrieve google bucket meta data, dbGap information and gen3 meta associated with terra workspace."""
import logging

from anvil.terra.workspace import workspace_factory
from anvil.terra.api import get_projects
# from anvil.cache import memoize
from collections import defaultdict
from attrdict import AttrDict

import sqlite3
import pickle
import hashlib
import json


class Reconciler():
    """Retrieve google bucket meta data, dbGap information and gen3 meta associated with terra workspace."""

    def __init__(self, name, user_project, namespaces, project_pattern, drs_file_path, terra_output_path, data_ingestion_tracker):
        """Initialize properties, set id to namespaces/project_pattern."""
        self.name = name
        self._user_project = user_project
        self.namespaces = namespaces
        self.project_pattern = project_pattern
        self._workspaces = None
        self._logger = logging.getLogger(__name__)
        self.id = f"{namespaces}/{project_pattern}"
        self.drs_file_path = drs_file_path
        self.terra_output_path = terra_output_path
        self.data_ingestion_tracker = data_ingestion_tracker

    @property
    def workspaces(self):
        """Terra workspaces that match namespace & project_pattern."""
        if not self._workspaces:
            self._workspaces = [
                workspace_factory(w, user_project=self._user_project, drs_file_path=self.drs_file_path, data_ingestion_tracker=self.data_ingestion_tracker)
                for w in get_projects(self.namespaces, self.project_pattern)
            ]
            for w in self._workspaces:
                w.attributes['reconciler_name'] = self.name

        return self._workspaces

    def make_hash(self, d):
        """Return the hash of a dictionary."""
        d_hash = hashlib.md5()
        # We need to sort arguments so {'a': 1, 'b': 2} is
        # the same as {'b': 2, 'a': 1}
        encoded = json.dumps(d, sort_keys=True).encode()
        d_hash.update(encoded)
        return d_hash.hexdigest()

    def reconcile_schemas(self):
        """Report workspaces that share common entities. Sorted in descending order of number of matching workspaces."""
        # sort workspaces into those who share a common schema
        def workspace_len(e):
            return len(e[1])

        workspace_entities = defaultdict(list)
        for w in self.workspaces:
            if len(w.schemas.keys()) == 0:
                workspace_entities[""].append(w.attributes.workspace.name)
                continue
            entity_list = ",".join(sorted(w.schemas.keys()))
            workspace_entities[entity_list].append(w.attributes.workspace.name)

        sorted_workspaces = sorted([(entities, workspaces) for (entities, workspaces) in workspace_entities.items()], key=workspace_len, reverse=True)
        consensus_schema = None
        for w in self.workspaces:
            if len(w.schemas.keys()) == 0:
                continue
            if w.attributes.workspace.name == sorted_workspaces[0][1][0]:
                consensus_schema = w.schemas
                break

        conformant = {'entities': [], 'workspaces': []}
        if len(sorted_workspaces) > 0:
            conformant = {
                'entities': sorted_workspaces[0][0].split(","),
                'workspaces': sorted_workspaces[0][1]
            }
        reconciled_schemas = {
            'conformant': conformant,
            'incompatible': [{'entities': rs[0], 'workspaces': rs[1]} for rs in sorted_workspaces[1:]],
            'schema_conflict_sample': [],
            'schema_conflict_subject': [],
            'missing_schema': [],
            'consensus_schema': consensus_schema
        }
        reconciled_schemas = AttrDict(reconciled_schemas)
        # print(datetime.datetime.now(), 'sorted_schemas')

        # check subjects and samples to see that data conforms to schema
        for name in reconciled_schemas.conformant.workspaces:
            for w in self.workspaces:
                if name == w.attributes.workspace.name:
                    if w.subject_property_name not in w.schemas:
                        reconciled_schemas['missing_schema'].append(name)
                        self._logger.debug(f"{w.name} {w.subject_property_name} not found in schema {w.schemas}.")
                        continue
                        # raise Exception(f"{w.name} {w.subject_property_name} not found in schema {w.schemas}.")
                    if len(w.subjects) == 0:
                        self._logger.debug(f"{w.name} has no subjects")
                        continue                        
                    if sorted(w.schemas[w.subject_property_name]['attributeNames']) != sorted(w.subjects[0].attributes.keys()):
                        reconciled_schemas['schema_conflict_subject'].append(name)
                        self._logger.debug(f"{w.name} schema_conflict due to subject.")
                    if 'sample' not in w.schemas:
                        reconciled_schemas['schema_conflict_sample'].append(name)
                        continue
                    if len(w.samples) == 0:
                        reconciled_schemas['schema_conflict_sample'].append(name)
                        continue
                    if 'attributes' not in w.samples[0]:
                        reconciled_schemas['schema_conflict_sample'].append(name)
                        continue
                    if sorted(w.schemas['sample']['attributeNames']) != sorted(w.samples[0].attributes.keys()):
                        reconciled_schemas['schema_conflict_sample'].append(name)
                        self._logger.debug(f"{w.name} schema_conflict due to sample.")

                    if w.subject_property_name not in w.schemas:
                        self._logger.debug(f"ERROR {w.name} no {w.subject_property_name} in schema? {w.schema}.")

        conformant_workspace_names = list(set(reconciled_schemas.conformant.workspaces))

        for name in reconciled_schemas.schema_conflict_sample:
            if name in conformant_workspace_names:
                self._logger.debug(f"removing schema_conflict {name}")
                conformant_workspace_names.remove(name)
        for name in reconciled_schemas.schema_conflict_subject:
            if name in conformant_workspace_names:
                self._logger.debug(f"removing schema_conflict {name}")
                conformant_workspace_names.remove(name)

        # print(datetime.datetime.now(), 'reconciled_schemas')

        return AttrDict(reconciled_schemas)

    @property
    def problems(self):
        """Aggregate all problems."""
        workspaces = self.workspaces
        _problems = {
            'inconsistent_entityName': [w.name for w in workspaces if w.inconsistent_entityName],
            'inconsistent_subject': [w.name for w in workspaces if w.inconsistent_subject],
            'missing_blobs': [w.name for w in workspaces if w.missing_blobs],
            'missing_samples': [w.name for w in workspaces if w.missing_samples],
            'missing_project_files': [w.name for w in workspaces if w.missing_project_files],
            'missing_subjects': [w.name for w in workspaces if w.missing_subjects],
        }
        return [k for k, v in _problems.items() if v]

    @property
    def blob_sizes(self):
        """Aggregate blob sizes by property name."""
        _blob_sizes = defaultdict(int)
        for w in self.workspaces:
            for k, v in w.blob_sizes.items():
                _blob_sizes[k] += v
        return _blob_sizes

    @property
    def project_file_blob_sizes(self):
        """Aggregate blob sizes by unique blob name."""
        _blob_sizes = {}
        for w in self.workspaces:
            # print(w.name)
            if w.project_files:
                for k, v in w.project_files.items():
                    # print(w.name)
                    # print(k, v)
                    _blob_sizes[v.value] = v.blob.size
        return _blob_sizes

    @property
    def dashboard_views(self):
        """Aggregate dashboard views."""
        for w in self.workspaces:
            v = w.dashboard_view
            v['problems'] = [k for k, v in v['problems'].items() if v]
            v['source'] = self.name
            yield v

    def save(self):
        """Persist workspaces to db."""
        entities = Entities(terra_output_path=self.terra_output_path, user_project=self._user_project)
        names = []
        for w in self.workspaces:
            if w.name in names:
                raise Exception(w.name)
            names.append(w.name)
            entities.save(w)
        entities.index()


class Entities:
    """Represent workspace objects."""

    def __init__(self, terra_output_path, user_project):
        """Simplify blob."""
        # self.avro_path = avro_path
        self.terra_output_path = terra_output_path
        self._conn = sqlite3.connect(self.terra_output_path)
        self.user_project = user_project
        cur = self._conn.cursor()
        cur.executescript("""
        CREATE TABLE IF NOT EXISTS vertices (
            key text PRIMARY KEY,
            submitter_id text,
            name text,
            pickle text NOT NULL
        );
        CREATE TABLE IF NOT EXISTS edges (
            src text,
            dst text,
            src_name text,
            dst_name text
        );
        """)
        self._conn.commit()
        # optimize for single thread speed
        self._conn.execute('PRAGMA synchronous = OFF')
        self._conn.execute('PRAGMA journal_mode = OFF')
        self._conn.commit()
        self._conn.close()
        self._conn = sqlite3.connect(self.terra_output_path, check_same_thread=False, isolation_level='DEFERRED')

    def put(self, key, submitter_id, name, data, cur):
        """Save an item."""
        cur.execute("REPLACE into vertices values (?, ?, ?, ?);", (key, submitter_id, name, pickle.dumps(data)))
        # self._logger.debug(f"put {key}")

    def get(self, key=None):
        """Retrieve an item."""
        cur = self._conn.cursor()
        data = cur.execute("SELECT pickle FROM vertices where key=?", (key,)).fetchone()
        if data:
            v = pickle.loads(data[0])
            src_edges = [(d[0], d[1]) for d in cur.execute("SELECT src, src_name FROM edges where dst=?", (key,)).fetchall()]
            edges = {}
            for src in src_edges:
                if src[1] not in edges:
                    edges[src[1]] = []
                for d in cur.execute("SELECT pickle, key, name FROM vertices where key=?", (src[0],)).fetchall():
                    edges[src[1]].append(pickle.loads(d[0]))

            return {'vertex': v, 'edges': edges}
        assert False, f"NOT FOUND {key}"

    def get_by_name(self, name=None):
        """Retrieve all items with name."""
        cur = self._conn.cursor()
        data = cur.execute("SELECT pickle FROM vertices where name=?", (name,)).fetchall()
        return [pickle.loads(d[0]) for d in data]

    def put_edge(self, src, dst, src_name, dst_name, cur):
        """Save edge."""
        cur.execute("REPLACE into edges values (?, ?, ?, ?);", (src, dst, src_name, dst_name))

    def save(self, workspace):
        """Load sqlite db from workspace."""
        cur = self._conn.cursor()
        logging.getLogger(__name__).debug(f'Starting save {workspace.name}')
        self.put(workspace.id, workspace.name, 'workspace', workspace, cur)
        for subject in workspace.subjects:
            subject_id = f"{workspace.name}/Subject/{subject.id}"
            self.put(subject_id, subject.id, 'subject', subject, cur)
            self.put_edge(subject_id, workspace.id, 'subject', 'workspace', cur)
            for sample in subject.samples:
                sample_id = f"{workspace.name}/Sample/{sample.id}"
                self.put(sample_id, sample.id, 'sample', sample, cur)
                self.put_edge(sample_id, subject_id, 'sample', 'subject', cur)
                for blob_id, blob in sample.blobs.items():
                    self.put(blob_id, blob_id, 'blob', blob, cur)
                    self.put_edge(blob_id, sample_id, 'blob', 'sample', cur)
                    if 'ga4gh_drs_uri' in blob:
                        drs = {'uri': blob['ga4gh_drs_uri']}
                        self.put(drs['uri'], sample.id, 'drs', drs, cur)
                        self.put_edge(drs['uri'], sample_id, 'drs', 'sample', cur)
        logging.getLogger(__name__).info(f'Saved {workspace.name}')
        self._conn.commit()

    def index(self):
        """Index the vertices and edges."""
        logging.getLogger(__name__).info('Indexing vertices & edges')
        cur = self._conn.cursor()
        cur.executescript("""
        CREATE  INDEX IF NOT EXISTS vertices_submitter_id ON vertices(submitter_id);
        CREATE UNIQUE INDEX IF NOT EXISTS edges_src_dst ON edges(src, dst, src_name, dst_name);
        CREATE  INDEX IF NOT EXISTS edges_dst ON edges(dst);
        CREATE INDEX IF NOT EXISTS vertices_name ON vertices(name)
        """)
        self._conn.commit()



#!/usr/bin/env python3

"""Read metadata from terra workspaces."""

import json
import os.path

from anvil.etl.transformers import _recursive_default_dict
from anvil.etl.utilities.entities import Entities
from pathlib import Path
import pickle

# from . import LogCapture
# import logging
# logger = LogCapture('anvil.etl.transformers.normalizer')
# logger.addHandler(logging.root.handlers[0])

from . import logger

from anvil.etl.transformers.normalizer_methods import *
from .fhir_writer import ensure_data_store_name


def _extract_ontology_fields(output_path=None, entities=None, workspace_name=None):
    """Query db, sample first row in all entities, determine fields that have references to ontology."""
    if not entities:
        assert output_path
        entities = Entities(path=f"{output_path}/terra_entities.sqlite")
    for workspace in entities.get_by_label('workspace'):
        workspace = AttrDict(workspace)
        if workspace_name and workspace.workspace.name != workspace_name:
            continue
        schema = AttrDict(entities.get_edges_by_label(src=workspace.workspace.name, src_name='workspace', dst_name='schema'))
        for entity_name in schema.schema.keys():
            if entity_name == 'schema':
                continue
            ontology_keys = set()

            children = entities.get_edges_by_label(src=workspace.workspace.name, src_name='workspace', dst_name=entity_name, limit=20)
            for entity_name, child in children.items():

                for k, v in child['attributes'].items():
                    if isinstance(v, str) and 'gs://' not in v and ':' in v and 'date' not in k and 'pfb' not in k:
                        ontology_keys.add(k)

            if len(ontology_keys) == 0:
                continue
            yield {
                'consortium_name': workspace.consortium_name,
                'workspace_name': workspace.workspace.name,
                'entity_name': entity_name,
                'ontology_fields': list(ontology_keys),
            }


def _normalize_workspace(consortium_name, workspace, config, entities, output_path, validate_buckets):
    """Stitch the graph together."""

    #
    # Uses command pattern to wrangle some very messy data
    #

    #
    # 'main'
    #
    if list(workspace.children.keys()) == ['schema']:
        logger.error(('no.child.entities', consortium_name, workspace.workspace.name))
        return None

    # pass context to commands
    context = AttrDict({'logged_already': [], 'output_path': output_path, 'entities': entities})

    # find config
    exec_command(context, consortium_name, workspace, config, 'consortium_config')
    assert context.consortium_config, "Should set context config"
    assert context.inverted_config, "Should set context inverted_config"

    # who ran this study?
    exec_command(context, consortium_name, workspace, config, 'practitioner')
    assert workspace.practitioner, (f"Should set workspace.practitioner `{workspace.practitioner}`", workspace.workspace.attributes)

    # find entities with bucket fields
    exec_command(context, consortium_name, workspace, config, 'ensure_bucket_fields')
    if not context['bucket_fields']:
        logger.warning(("no.entity.contains.bucket.objects", consortium_name, workspace.workspace.name))

    # deduce entity names in workspace
    exec_command(context, consortium_name, workspace, config, 'patient_entity_names')
    assert context.patient_entity_names, "Should set context patient_entity_names"

    exec_command(context, consortium_name, workspace, config, 'specimen_entity_name')
    assert context.specimen_entity_name, "Should set context specimen_entity_name"
    assert isinstance(context.specimen_entity_name, str), context.specimen_entity_name

    # traverse from specimen to patient
    exec_command(context, consortium_name, workspace, config, 'link_specimen_to_patient')
    if 'specimen_patient_ids' not in workspace:
        logger.error(("no.specimen_patient_ids", consortium_name, workspace.workspace.name))
        return
    for specimen_id, patient_id in workspace.specimen_patient_ids:
        assert specimen_id and patient_id, f"All specimens should have patient: {(specimen_id, patient_id)}"

    # find tasks
    exec_command(context, consortium_name, workspace, config, 'task_entity_names')
    assert context.task_entity_names, "Should set context.task_entity_names"

    exec_command(context, consortium_name, workspace, config, 'ensure_tasks')
    assert workspace.tasks, ("context.tasks.empty", context.keys(), context.task_entity_names, context.consortium_config, workspace.children.keys())
    assert len(workspace.tasks) > 0, "context.tasks should be > 0"

    exec_command(context, consortium_name, workspace, config, 'ensure_tasks_populated')
    assert workspace.tasks, ("context.tasks.empty", context.keys(), context.task_entity_names, workspace.children.keys())
    assert len(workspace.tasks) > 0, "context.tasks should be > 0"

    exec_command(context, consortium_name, workspace, config, 'ensure_tasks_linked_to_documents')
    if len(workspace.tasks) == 0:
        logger.warning(("no.tasks", consortium_name, workspace.workspace.name))
    else:
        for task_source, task_list in workspace.tasks.items():
            for task in task_list:
                assert len(task['inputs']) > 0, ("tasks.missing.inputs", workspace.workspace.name, task_source, task, context.bucket_fields)
                assert len(task['outputs']) > 0, ("tasks.missing.outputs", workspace.workspace.name, task_source, task, context.bucket_fields)

    # link bucket data
    context['validate_buckets'] = validate_buckets
    exec_command(context, consortium_name, workspace, config, 'blob_attributes')

    # setup a normalized patient model
    # patient has specimens
    # specimens have tasks
    # tasks have input = specimen  
    # tasks have output = [document_reference]
    exec_command(context, consortium_name, workspace, config, 'patient_model')
    if 'patients' not in workspace:
        logger.error(('no.patients', workspace.workspace.name,))


def ontologies(output_path, consortium_name, workspace_name, details):
    """Determine if there are any fields that are ontologies in workspace."""
    for workspace_name in fetch_workspace_names(output_path=output_path, requested_consortium_name=consortium_name, workspace_name=workspace_name):
        yield [ontology_field for ontology_field in _extract_ontology_fields(output_path=output_path, workspace_name=workspace_name)]


def get_pickled_workspace(output_path, consortium_name, workspace_name):
    """Unpickle the normalized workspace."""
    path = f"{output_path}/workspaces/{consortium_name}/{workspace_name}.pickle"
    if not os.path.exists(path):
        return None
    try:
        with open(path, 'rb') as input_stream:
            return pickle.load(input_stream)
    except Exception as e:
        logger.warning(("no.pickle", workspace_name, str(e)))


def write_workspace(output_path, consortium_name, workspace_name, validate_buckets, config):
    """Read workspaces from db, normalize and write to file system."""

    for consortium_name, workspace in normalize(output_path, consortium_name, workspace_name, validate_buckets, config=config):
        Path(f"{output_path}/workspaces/{consortium_name}").mkdir(parents=True, exist_ok=True)
        with open(f"{output_path}/workspaces/{consortium_name}/{workspace.workspace.name}.pickle", 'wb') as output_stream:
            pickle.dump(workspace, output_stream)


def analyze(output_path, consortium_name, workspace_name, details, validate_buckets, config):
    """Read workspaces from db, normalize and analyze."""
    assert consortium_name
    G = _recursive_default_dict()
    for consortium_name, workspace in normalize(output_path, consortium_name, workspace_name, validate_buckets, config):
        specimen_count = 0
        task_count = 0
        patient_count = 0
        uniq_documents = set()
        documents_size = 0
        uniq_genders = set()
        phenotype_values = defaultdict(set)
        disease_values = defaultdict(set)
        family_relationship_keys = set()
        family_relationship_items = {}
        files = {}
        data_store_name = "None"

        for p in workspace.patients.values():
            patient_count += 1
            if p['gender']:
                uniq_genders.add(p['gender'])
            assert isinstance(p['phenotype_values'], list), (p['phenotype_values'], workspace_name)
            assert isinstance(p['disease_values'], list), (p['disease_values'], workspace_name)
            for phenotype_value in p['phenotype_values']:
                for key, value in phenotype_value.items():
                    # add list of values to set
                    phenotype_values[key].update(value)
            for disease_value in p['disease_values']:
                for key, value in disease_value.items():
                    disease_values[key].update(value)
            for family_relationship_value in p['family_relationship_values']:
                for key, value in family_relationship_value.items():
                    family_relationship_keys.add(key)
                family_relationship_items[p['name']] = family_relationship_value

            if 'specimens' not in p:
                continue

            specimen_count += len(p['specimens'])
            for s in p['specimens']:
                if 'tasks' in s:
                    task_count += len(s['tasks'])
                    for task in s['tasks']:
                        for output_source, output in task['outputs'].items():
                            for output_property, blob in output.items():
                                if blob['url'] not in uniq_documents:
                                    documents_size += int(blob.get('size', 0))
                                    uniq_documents.add(blob['url'])
                                    extension = blob['url'].split('/')[-1].replace('.gz', '').split('.')[-1]
                                    if extension not in files:
                                        files[extension] = {'size': 0, 'count': 0, 'drs_count': 0}
                                    files[extension]['size'] += int(blob.get('size', 0))
                                    files[extension]['count'] += 1
                                    if 'drs_uri' in blob and blob['drs_uri']:
                                        files[extension]['drs_count'] += 1

        # xform to a list
        for k, v in phenotype_values.items():
            phenotype_values[k] = sorted(v)
        for k, v in disease_values.items():
            disease_values[k] = sorted(v)

        G[consortium_name][workspace.workspace.name]['consortium'] = consortium_name
        G[consortium_name][workspace.workspace.name]['workspace'] = workspace.workspace.name
        G[consortium_name][workspace.workspace.name]['nodes']['patients'] = patient_count
        G[consortium_name][workspace.workspace.name]['nodes']['specimens'] = specimen_count
        G[consortium_name][workspace.workspace.name]['nodes']['tasks'] = task_count
        G[consortium_name][workspace.workspace.name]['nodes']['documents'] = len(uniq_documents)
        G[consortium_name][workspace.workspace.name]['phenotype_values'] = phenotype_values
        G[consortium_name][workspace.workspace.name]['disease_values'] = disease_values
        G[consortium_name][workspace.workspace.name]['gender'] = list(uniq_genders)
        G[consortium_name][workspace.workspace.name]['family_relationship_keys'] = list(family_relationship_keys)
        G[consortium_name][workspace.workspace.name]['documents_size'] = documents_size
        G[consortium_name][workspace.workspace.name]['files'] = files
        G[consortium_name][workspace.workspace.name]['data_store_name'] = ensure_data_store_name(workspace)
        G[consortium_name][workspace.workspace.name]['indication'] = None
        if 'tracker' in workspace and workspace['tracker']:
            G[consortium_name][workspace.workspace.name]['indication'] = workspace['tracker']["library:indication"]

        error_count = 0
        for error_name, entries in workspace.errors.items():
            error_count += len(entries)
            if details:
                G[consortium_name][workspace.workspace.name]['errors'][error_name] = entries
            else:
                G[consortium_name][workspace.workspace.name]['errors'][error_name] = len(entries)
        G[consortium_name][workspace.workspace.name]['error_count'] = error_count

    return G


def qa(output_path, consortium_name, workspace_name, config):
    """Produce a QA report."""
    pass


def fetch_workspace_names(output_path, requested_consortium_name, workspace_name):
    """Query db and return workspace names."""
    workspace_names = []
    entities = Entities(path=f"{output_path}/terra_entities.sqlite")
    for workspace in entities.get_by_label('workspace'):
        workspace = AttrDict(workspace)
        # filter consortium_name
        if requested_consortium_name and requested_consortium_name != workspace.consortium_name:
            continue

        # filter workspace_name
        if workspace_name and workspace.workspace.name != workspace_name:
            continue

        workspace_names.append((workspace.consortium_name, workspace.workspace.name))

    return workspace_names


def normalize(output_path, requested_consortium_name, workspace_name, validate_buckets, config):
    """Read workspaces from db, normalize."""
    workspace = get_pickled_workspace(output_path, requested_consortium_name, workspace_name)
    if workspace:
        yield workspace.consortium_name, workspace
        return

    config = config['consortiums']

    entities = Entities(path=f"{output_path}/terra_entities.sqlite")

    tracker = json.load(open(f"{output_path}/data_ingestion_tracker.json"))

    for workspace in entities.get_by_label('workspace'):
        workspace = AttrDict(workspace)
        # filter consortium_name
        if requested_consortium_name and requested_consortium_name != workspace.consortium_name:
            continue

        consortium_name = workspace.consortium_name
        # filter workspace_name
        if workspace_name and workspace.workspace.name != workspace_name:
            continue

        logger.info((consortium_name, workspace.workspace.name))

        # get all the other entities for this workspace
        children = entities.get_edges(src=workspace.workspace.name, src_name='workspace')
        workspace.children = children

        # clear logger's entries so we can harvest later
        logger.clear()
        # wrangle the entities into a consistent, predictable form
        _normalize_workspace(consortium_name, workspace, config, entities, output_path, validate_buckets)
        try:
            assert 'patients' in workspace , 'missing.patients'
            # assert workspace.specimens, 'missing.specimens'
            # assert workspace.specimen_patient_ids, 'missing.specimen.to.patient.links'
            # assert workspace.tasks, 'missing.tasks'
            # for task_entity in workspace.tasks:
            #     for task in workspace.tasks[task_entity]:
            #         for input in task['inputs']:
            #             assert 'fhir_entity' in input, 'missing.fhir_entity.in.task'
            #         for output_source, output in task['outputs'].items():
            #             for output_property, blob in output.items():
            #                 assert 'url' in blob, ('missing.url.in.blob', output_source, output_property)
        except Exception as e:
            logger.error((str(e), consortium_name, workspace.workspace.name))
            continue

        tracker_info = [p for p in tracker if p['name'] == workspace.workspace.name]
        workspace.tracker = next(iter(tracker_info), None)

        workspace.errors = {k.replace('.', '-'): v for k, v in logger.entries.items()}
        # # workspace.errors = [v for v in logger.entries.values()]
        # for v in logger.entries.values():
        #     for item in v:
        #         print(item)
        # workspace.errors = [list(v) for v in logger.entries.values()]

        yield consortium_name, workspace




"""Represent fhir entity."""

from anvil.transformers.fhir.research_study import ResearchStudy
from anvil.transformers.fhir.specimen import Specimen
from anvil.transformers.fhir import make_id
from anvil.transformers.fhir.patient import Patient


class Task:
    """Create fhir entity."""

    class_name = "task"
    resource_type = "Task"

    @staticmethod
    def build_entity():
        """Create fhir entity."""
        pass


class SpecimenTask:
    """Create fhir entity linking Specimen to Blobs."""

    class_name = "task"
    resource_type = "Task"

    @staticmethod
    def slug(specimen):
        """Make id."""
        return make_id("Task", specimen.id)

    @staticmethod
    def build_entity(inputs, outputs, subject):
        """Create fhir entity."""
        specimen = inputs[0]
        inputs = [
            {
                "type": {
                    "coding": [
                        {
                            "code": Specimen.resource_type
                        }
                    ]
                },
                "valueReference": {
                    "reference": f"{Specimen.resource_type}/{Specimen.slug(specimen)}"
                }
            }
        ]
        # inject this task context into DocumentReferences
        for blob in outputs:
            blob['context'] = {
                "related": [
                    {
                        "reference": f"Task/{SpecimenTask.slug(specimen)}"
                    }
                ]
            }

        outputs = [
            {
                "type": {
                    "coding": [
                        {
                            "code": blob['resourceType']
                        }
                    ]
                },
                "valueReference": {
                    "reference": f"{blob['resourceType']}/{blob['id']}"
                }
            }
            for blob in outputs]

        return {
            "resourceType": "Task",
            "id": SpecimenTask.slug(specimen),
            "meta": {
                "profile": [
                    "https://ncpi-fhir.github.io/ncpi-fhir-ig/StructureDefinition/ncpi-specimen-task"
                ]
            },
            "identifier": [
                {
                    "system": f"https://anvil.terra.bio/#workspaces/anvil-datastorage/{subject.workspace_name}",
                    "value": f"{specimen.id}/Task/AnVILInjest",
                },
                {
                    "system": "urn:ncpi:unique-string",
                    "value": f"{subject.workspace_name}/Patient/{subject.id}/Specimen/{specimen.id}/Task/AnVILInjest",
                },
            ],
            "status": "accepted",
            "intent": "unknown",
            "input": inputs,
            "output": outputs,
            "focus": {
                "reference": f"{Specimen.resource_type}/{Specimen.slug(specimen)}"
            },
            "for": {
                "reference": f"{Patient.resource_type}/{Patient.slug(subject)}"
            },
            "owner": {
                "reference": f"Organization/{ResearchStudy.slug(specimen)}"
            }
        }



"""Represent fhir entity."""

from anvil.transformers.fhir.patient import Patient
from anvil.transformers.fhir.research_study import ResearchStudy


class ResearchSubject:
    """Create fhir entity."""

    class_name = "research_subject"
    resource_type = "ResearchSubject"
    # Patient already uses CONCEPT.PARTICIPANT.TARGET_SERVICE_ID,
    # so the below is set to None
    target_id_concept = None

    @staticmethod
    def build_entity(subject):
        """Create fhir entity."""
        study_slug = ResearchStudy.slug(subject)
        patient_slug = Patient.slug(subject)
        research_subject_status = "on-study"  # QUESTION: https://www.hl7.org/fhir/valueset-research-subject-status.html

        entity = {
            "resourceType": ResearchSubject.resource_type,
            "id": patient_slug,
            "meta": {
                "profile": [
                    "http://hl7.org/fhir/StructureDefinition/ResearchSubject"
                ]
            },
            "identifier": [
                {
                    "system": f"https://anvil.terra.bio/#workspaces/anvil-datastorage/{subject.workspace_name}",
                    "value": subject.id,
                },
                {
                    "system": "urn:ncpi:unique-string",
                    "value": f"{subject.workspace_name}/ResearchSubject/{subject.id}",
                },
            ],
            "status": research_subject_status,
            "study": {
                "reference": f"ResearchStudy/{study_slug}"
            },
            "individual": {
                "reference": f"Patient/{patient_slug}"
            },
        }

        return entity
